/*
   pid.c -- PIKT routines to find pid's and ppid's

   $Header: /var/cvs/PIKT/pikt/src/lib/pid.c,v 1.12 2004/01/29 15:06:08 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   I strongly suspect that there are better ways to do this!
   Suggestions, anyone?  Help!

*/

#include "common.h"

/* for the given process, return either its PID or PPID

   we used to invoke ioctl() on each /proc/<pid> to get the process names,
   but we saw several cases where the resultant process name for a given
   pid didn't match the process name for that pid as shown in the ps display;
   the technique of reading the ps display, although perhaps less elegant,
   gives us the consistent results we desire
 */
pid_t
procid (char *procname, int type)
{
	static char fnnam[] = "procid";
	pid_t pspid;
	long pid, ppid;
	char *cmd;
	int procnamelen;
	#if PSPARSE == 1
	char psp[MAXLINE+1];
	#endif
	char *psproc;
	char line[MAXLINE+1];
	FILE *ps;
	char *cp;

	pspid = 0;
	procnamelen = strlen(procname);
	cmd = (char *) mymalloc(strlen(pscmd)+strlen(PSOPTS)+2, fnnam);
	sprintf(cmd, "%s %s", pscmd, PSOPTS);
	if ((ps = popen(cmd, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, FALSE, "popen() error");
		free(cmd);
		return ERROR;
	}
	free(cmd);
	if (fgets(line, MAXLINE+1, ps) == NULL) {	/* throw away header */
		if (ferror(ps))
			logreturn(PIKTLOG, LOG_ERR, ERRID_FGETS_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fgets() error for %s", ps);
	}
	while (fgets(line, MAXLINE+1, ps)) {
		*strchr(line, '\n') = '\0';
		#if PSPARSE == 1
		if (sscanf(line, "%ld %ld %s", &pid, &ppid, psp) != 3) {
			logreturn(PIKTLOG, LOG_WARNING, ERRID_PS_READ_ERROR, __FILE__, __LINE__, fnnam, FALSE, "error reading pid/ppid/psproc for line %s", line);
			continue;
		}
		/* strip away path, if any, leaving only basename */
		if ((cp = strrchr(psp, '/')))
			psproc = cp+1;
		else
			psproc = psp;
		#else
		# if PSPARSE == 2
		if (sscanf(line, "%*s %*s %ld %ld %*s %*s %*s %*s %*s %*s %*s %*s %*s", &pid, &ppid) != 2) {
			logreturn(PIKTLOG, LOG_WARNING, ERRID_PS_READ_ERROR, __FILE__, __LINE__, fnnam, FALSE, "error reading pid/ppid for line %s", line);
			continue;
		}
		if ((psproc = strrchr(line, ':')))
			psproc += 4;
		else {
			logreturn(PIKTLOG, LOG_WARNING, ERRID_PS_READ_ERROR, __FILE__, __LINE__, fnnam, FALSE, "error reading psproc for line %s", line);
			continue;
		}
		# else
		#  if PSPARSE == 3
		if (sscanf(line, "%*s %*s %*s %ld %ld %*s %*s %*s %*s %*s %*s %*s %*s %*s", &pid, &ppid) != 2) {
			logreturn(PIKTLOG, LOG_WARNING, ERRID_PS_READ_ERROR, __FILE__, __LINE__, fnnam, FALSE, "error reading pid/ppid for line %s", line);
			continue;
		}
		if ((psproc = strrchr(line, ':')))
			psproc += 4;
		else {
			logreturn(PIKTLOG, LOG_WARNING, ERRID_PS_READ_ERROR, __FILE__, __LINE__, fnnam, FALSE, "error reading psproc for line %s", line);
			continue;
		}
		#  else
		#   include "Bletch: How do we parse `ps' output?"
		#  endif
		# endif
		#endif
		/* for some strange reason, linux ps appends a space to the
		   end of each ps line
		 */
		if (! strncmp(psproc, procname, (size_t)procnamelen) && ((*(psproc+procnamelen) == '\0') || (*(psproc+procnamelen) == ' '))) {
			if (strcmp(procname, "piktd") || (ppid == 1)) {
				if (pspid == 0) {
					switch (type) {
						case PROCID:
							pspid = (pid_t) pid;
							break;
						case PPROCID:
							pspid = (pid_t) ppid;
							break;
					}
				} else {
					logreturn(PIKTLOG, strcmp(procname, "piktd") ? LOG_WARNING : LOG_ERR, ERRID_MULTIPLE_PROCS, __FILE__, __LINE__, fnnam, FALSE, "failed to determine unique pid for proc %s: multiple processes running", procname);
					pclose(ps);
					return ERROR;
				}
			}
		}
	}
	pclose(ps);
	if ((pspid == 0)
		#if defined(SKIP_PROC)
		&& strcmp(procname, SKIP_PROC)
		#endif
						) {
		if (! strcmp(procname, "piktd"))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_PROC_NOT_RUNNING, __FILE__, __LINE__, fnnam, FALSE, "failed to determine pid for proc %s: process not running", procname);
		return NIL;
	} else
		return pspid;
}
